/* floorcontrol.cpp : file responsible for the conference control functions
 * (c) 2003 Columbia University. All rights reserved.
 */
/**@name floorcontrol.cpp -- Floor control module for sipua.
 * This file contains the implementation for the floor control
 * module for sipua. It implements sending of various SOAP commands
 * and receiving events and results. 
 */
//@{

#include <iostream>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>

#include <string>

#include <list>        /* STL for list */
#include <map>         /* STL for map */
#include <vector>      /* STL for vector */

#include "soapH.h"     /* Floor control definitions */
#include "command.h"
#include "string-util.h"

/**
 * All the name spaces to be used by the gSOAP for floor control.
 */
#include "FloorControl.nsmap"
#include "soapFloorControlProxy.h"

using namespace std;

/* Debug trace with line number */
#define TRACE(c)  cout << "[" << __LINE__ << "] " << c << endl

static void display_floorcontrol_help(const std::string topic = "")
{
  cout << 
    "floor command [param ...]\n"
    " floor is used to issue floor control commands.\n"
    " command is one of the following:\n"
    "    create <floor>        - to create a floor\n"
    "    remove <resources>    - to remove a floor\n"
    "    change <floor>        - to change a floor\n"
    "    claim  <claim>        - to claim a floor\n"
    "    release <holding>     - to release a floor\n"
    "    grant <holding>       - to grant a claim\n"
    "    revoke <holding>      - to revoke a floor\n"
    "    removeclaims <claim>  - to remove a claim\n"
    "    reorder <claim> <operation> - to reorder a claim\n"
    " The parameters <floor>, <resources>, <claim>, <holding>\n"
    " and <operation> are defined as follows with examples.\n"
    " [ ] means optional.\n"
    "  <floor>     = [mid:1,... [user1,...  [moderator1,...]]]\n"
    "  <resources> = mid:1,mid:2,...\n"
    "  <claim>     = [claimid[,user[,subject]]]\n"
    "  <holding>   = user1,user2,... [mid:1,mid:2,...]\n"
    "  <operation> = (up:1 | down:2 | top | bottom)\n"
    "here user should be same as the user@host used to identify a user\n"
       << flush;
}

static void parse_floor(_m__floorType & floor, CommandObject& c, int& count)
{
  floor.maxHolders = 1;
  floor.resources = NULL;
  floor.users = NULL;
  floor.moderators = NULL;
  
  list<string> r = strtokenize(c.GetArg(++count), ",");
  if (r.size() > 0) {
    floor.resources = new _m__resourcesType;
    for (list<string>::iterator i = r.begin(); 
         i != r.end(); ++i) {
      floor.resources->resource.push_back(*i);
    }
  }
  
  list<string> u = strtokenize(c.GetArg(++count), ",");
  if (u.size() > 0) {
    floor.users = new _m__usersType;
    for (list<string>::iterator i = u.begin(); 
         i != u.end(); ++i) {
      floor.users->user.push_back(*i);
    }
  }
  
  list<string> m = strtokenize(c.GetArg(++count), ",");
  if (m.size() > 0) {
    floor.moderators = new _m__moderatorsType;
    for (list<string>::iterator i = m.begin(); 
         i != m.end(); ++i) {
      floor.moderators->moderator.push_back(*i);
    } 
  }
}

static void parse_resources(_m__resourcesType& resources,
                            CommandObject& c, int& count)
{
  list<string> r = strtokenize(c.GetArg(++count), ",");
  for (list<string>::iterator i = r.begin(); i != r.end(); ++i) {
    resources.resource.push_back(*i);
  }
}

static void parse_claim(_m__claimType& claim, CommandObject& c, int& count)
{
  list<string> r = strtokenize(c.GetArg(++count), ",");
  list<string>::iterator i = r.begin();
  if (i != r.end()) {
    claim.claimID = *i;
    ++i;
  }
  if (i != r.end()) {
    claim.user = *i;
    ++i;
  }
  if (i != r.end()) {
    claim.subject = *i;
    ++i;
  }
  claim.resources = NULL;
}

static void parse_holding(_m__holdingType& holding, CommandObject& c,
                          int& count)
{
  list<string> u = strtokenize(c.GetArg(++count), ",");
  for (list<string>::iterator i = u.begin(); i != u.end(); ++i) {
    holding.users.user.push_back(*i);
  }

  list<string> r = strtokenize(c.GetArg(++count), ",");
  if (r.size() > 0) {
    holding.resources = new _m__resourcesType;
    for (list<string>::iterator i = r.begin(); i != r.end(); ++i) {
      holding.resources->resource.push_back(*i);
    }
  }
  else {
    /* Some how the GSOAP library expects this to be non-NULL. XXX
    * thats why we need to allocate it even when no resources was
    * specified.
    */
    holding.resources = new _m__resourcesType;
  }
}

static void parse_operation(_m__operationType& operation,
                            CommandObject& c, int& count)
{
  list<string> r = strtokenize(c.GetArg(++count), ":");
  list<string>::iterator i = r.begin();
  if (i != r.end()) {
    if ((*i) == "up") {
      operation.operator_ = up;
    }
    else if ((*i) == "down") {
      operation.operator_ = down;
    }
    else if ((*i) == "top") {
      operation.operator_ = top;
    }
    else if ((*i) == "bottom") {
      operation.operator_ = bottom;
    }

    ++i;
  }
  if (i != r.end()) {
    operation.argument = (int *)malloc(sizeof(int));
    *operation.argument = atoi((*i).c_str());

    ++i;
  }
}

void FC_HandleCommand(CommandObject& c, FloorControl& fc)
{
  int ret = SOAP_OK;
  bool result = true;

  if (c.GetArgsCount() < 1) {
    display_floorcontrol_help();
  }
  else {
    std::string command = c.GetArg(1);
    int count = 1;
    if (command == "create") {
      _m__floorType floor;
      parse_floor(floor, c, count);
      ret = fc.CreateFloor(floor, result);
    }
    else if (command == "remove") {
      _m__resourcesType resources;
      parse_resources(resources, c, count);
      ret = fc.RemoveFloor(resources, result);
    } 
    else if (command == "change") {
      _m__floorType floor;
      parse_floor(floor, c, count);
      ret = fc.ChangeConfig(floor, result);
    }
    else if (command == "claim") {
      _m__claimType claim;
      _m__claimsType claims;
      parse_claim(claim, c, count);
      claims.claim.push_back(claim);
      ret = fc.ClaimFloor(claims, result);
    }
    else if (command == "release") {
      _m__holdingType holding;
      parse_holding(holding, c, count);
      ret = fc.ReleaseFloor(holding, result);
    }
    else if (command == "grant") {
      _m__holdingType holding;
      parse_holding(holding, c, count);
      ret = fc.GrantFloor(holding, NULL, result);
    }
    else if (command == "revoke") {
      _m__holdingType holding;
      parse_holding(holding, c, count);
      ret = fc.RevokeFloor(holding, result);
    }
    else if (command == "removeclaims") {
      _m__claimType claim;
      _m__claimsType claims;
      parse_claim(claim, c, count);
      claims.claim.push_back(claim);
      ret = fc.ClaimFloor(claims, result);
    }
    else if (command == "reorder") {
      _m__claimType claim;
      parse_claim(claim, c, count);
      _m__operationType operation;
      parse_operation(operation, c, count);
      ret = fc.ReorderClaims(NULL, claim, operation, result);
    }
    else {
      display_floorcontrol_help(command);
      result = false;
    }

    if (ret != SOAP_OK) {
      // display the SOAP fault message on the stderr 
      soap_print_fault(fc.soap, stdout); 
    }
    else if (result == false) {
      cout << "floor " << command << " failed" << endl;
    }
    else {
      cout << "floor " << command << " successful" << endl;
    }
  }
}


//@}
