extern "C" {
#include <pthread.h>
#include "timespec.h"

#include "strdupn.h"

extern char *optarg;
}


#include <stdio.h>
#include <utility>
#include <iostream>
#include <sstream>

#include <string>
#include <map>
#include <list>

#include "string-util.h"
#include "bluetooth.h"
#include "radio.h"  /* Audio tool API */


#include "iptelsession.h"
#include "sipcallobj.h"
#include "sipregobj.h"
#include "siplibrary.h"
#include "sipsession.h"

using namespace std;


//////////////////////////////// Local class definitions /////////////////

/**
 * MySIPUserAgent is defined as a user agent or endpoint to handling
 * this SIP user agent.
 */
class MySIPUserAgent : public SIPSessionEndpoint
{
public:
  MySIPUserAgent();
  IptelCall *AllocateNewCallObject(const std::string callid);

  friend class MySIPCall;
};

/**
 * MySIPCall defines call back functions to handle this
 * user agents incoming calls, termination, rejection etc.
 */
class MySIPCall : public SIPSession
{
public:
  MySIPCall(MySIPUserAgent & ep, const std::string callid = "");

  ~MySIPCall();

  void   OnNewIncomingCall(IptelEndpoint & ep);
  bool   OnHangup();
  //void   OnCallEstablished();
  //bool   OnCallRejected();
  //bool   OnRemoteSessionChanged(const SessionDescription & sdp);
  bool   OnSuccessResponse(SIPResponseMessage & msg);
  bool   OnFailureResponse(SIPResponseMessage & msg);

  void   Publish(const std::string& event, const SIPMessageBody& body);

  void   UseAudio(const std::string name="");
  AudioTool * GetAudioTool() {return audio; }
  bool   SetRemoteAudioAddress(const AFIAddress & addr);
  AFIAddress GetSelfAudioAddress() const;
  bool   StartAudio();
  bool   StopAudio();

  friend int main(int argc, char *argv[]);

protected:
  /** Audio tool for this call */
  AudioTool * audio;
};

//////////////// Globals and file variables ////////////////////////////////

/**
 * Audio tool to use.
 */
static  std::string audio_tool;
static void handle_mapping_success(const std::string& mac_uri,
                                  const std::string& user);
static void handle_bluetooth();

static std::string proxy;
static std::string mapping_server;
static std::string presenceserver;
std::map<std::string, std::string> mac_users;
static void handle_device_add(const std::string& device);
static void handle_device_remove(const std::string& device);
static void query_mapping_server(const std::string& mac);
static void publish_user(const std::string& user, bool online);
static void register_user(const std::string& user,
                          const std::string& from,
                          bool online);

static void handle_mapping_success(const std::string& mac_uri,
                                  const std::string& user);


//////////////// Method defn for SIP ///////////////////////////////////////

MySIPUserAgent::MySIPUserAgent(): SIPSessionEndpoint()
{
}

IptelCall *MySIPUserAgent::AllocateNewCallObject(const std::string callid)
{
  /* Allocate a new SIP call */
  return new MySIPCall(*this, callid);
}

MySIPCall::MySIPCall(MySIPUserAgent & ep, const std::string callid)
: SIPSession(ep, callid)
{
  audio           = NULL;

  /* use the specified audio tool */
  UseAudio(audio_tool);
}

MySIPCall::~MySIPCall()
{
  if (audio != NULL) {
    audio->stop();
    delete audio;
  }
}

bool MySIPCall::OnSuccessResponse(SIPResponseMessage & msg)
{
  bool ret = SIPSession::OnSuccessResponse(msg);
  std::string method = msg.GetMethod();

  if (method == "PUBLISH") {
    SIPHeader to_hdr = msg.GetHeader("To");
    string to = to_hdr.GetValue();
    debug(DEBUG_APP, "OnSuccessResponse", "Successful PUBLISH to %s\n",
	  to.c_str());

    SIPHeader event_hdr = msg.GetHeader("Event");
    string event = event_hdr.GetValue();
    if (event == "presence") {
      /* presence server responded with success. ignore. */
    }

    Terminate();
    ret = true;
  }

  return ret;
}

bool MySIPCall::OnFailureResponse(SIPResponseMessage & msg)
{
  bool ret = SIPSession::OnFailureResponse(msg);
  int status = msg.GetStatusCode();
  std::string method = msg.GetMethod();
  cout << "OnFailure Response Called" << endl;
  //if (method == "PUBLISH") {
    //SIPHeader to_hdr = msg.GetHeader("To");
    //string to = to_hdr.GetValue();
    //error("OnFailureResponse", "Failed for PUBLISH to %s\n",
    //	  to.c_str());
    cout << " Response  " << status << msg.GetReasonPhrase() << endl;
    if(msg.Is3xxResponse()){
 
	    //SIPHeader event_hdr = msg.GetHeader("Event");
	    //string event = event_hdr.GetValue();
	    //cout << "Event is " << event << endl;
	    //if (event == "lookup") {
	      /* mapping server responded probably 302. 
		 so send publish and register now. */

	      SIPHeader f_adr =msg.GetHeader("From");
	      string mac_uri = f_adr.GetValue();
	      cout << "from address is"<< mac_uri << endl;

	      SIPHeader c_hdr = msg.GetHeader("Contact");
	      string c = c_hdr.GetValue();
	      cout << "New  contact is " << c << endl;
	      handle_mapping_success(mac_uri, c);
   	    //}
     }

    Terminate();
  //}

  return ret;
}


void MySIPCall::OnNewIncomingCall(IptelEndpoint & ep)
{
  MySIPUserAgent & e = (MySIPUserAgent &) endpoint;

  /* check the processing under "accept" command of sipua/main.cpp
   * and duplicate it here to accept the incoming call if the MAC 
   * address belongs to the local device. */
}

bool MySIPCall::OnHangup()
{
  StopAudio();

  SIPSession::OnHangup();
  return true;
}

void MySIPCall::UseAudio(const std::string name)
{
  if (audio != NULL) {
    audio->stop();
    delete audio;
    audio = NULL;
  }
  if (name == "") {
    /* Use the built in audio tool */
#ifdef WIN32
    audio = new WindowsAudioTool();
#else
    audio = new DeviceAudioTool();
#endif
  }
  else if (!strcasecmp(name.c_str(), "none")) {
    /* No audio */
    audio = NULL;
  }
  else if (!strcasecmp(name.c_str(), "loopback")) {
    /* Loopback audio */
    audio = new LoopbackTool();
  }
  else if (!strncasecmp(name.c_str(), "file://", 7)) {
    /* Have a file audio */
    audio = new FileAudioTool(name.substr(7, name.length() - 7), "");
  }
  else {
    /* Some application */
    audio = new ExternalAudioTool(name);
  }
}

bool  MySIPCall::SetRemoteAudioAddress(const AFIAddress & addr)
{
  return (audio != NULL) ? (audio->set_destination(addr), true) : false;
}

AFIAddress MySIPCall::GetSelfAudioAddress() const
{
  return (audio != NULL) ? audio->get_receive_address() : AFIAddress();
}

bool   MySIPCall::StartAudio()
{
  return (audio != NULL) ? audio->start() : false;
}

bool   MySIPCall::StopAudio()
{
  return (audio != NULL) ? audio->stop() : false;
}

static void set_local_sdp(MySIPCall & mycall, std::string payloadtype,
                          bool ipv6)
{
  /* 
   * Allocate self session info - here, since we dont know 
   * whether the caller is ipv6 capable we will put both
   * v4 and v6
   */
  SessionDescription *session;
  session = new SessionDescription;
  SessionMedia media;
  
  media.SetDirection("recv");
  media.SetDataType("audio");
  
  if (payloadtype.empty()) {
    payloadtype = "pcmu";
  }
  
  SessionFormat f(payloadtype);
  
  AFIAddress rxaddr = mycall.GetSelfAudioAddress();
  in_port_t port = rxaddr.get_port();

  if (port == 0) {
    error("set_self_sdp", "Audio port is not allocated\n");
  }
  
  /* Add IPv6 media */
  if (ipv6) {
    media.SetAddressType("IP6");
    rxaddr = SIPLibrary::GetHostIP6Addr();
    rxaddr.set_port(port);
    media.SetRxAddress(rxaddr); 
    media.AddFormat(f); 
    session->GetSessionList().push_back(media);
  }
  
  /* Add IPv6 media */
  media.SetAddressType("IP4");
  rxaddr = SIPLibrary::GetHostIPAddr();
  rxaddr.set_port(port);
  media.SetRxAddress(rxaddr);
  media.AddFormat(f);
  session->GetSessionList().push_back(media);

#ifdef USE_GSOAP
  std::cout << "Adding soap session" << std::endl;
  SessionMedia soap;
  SessionFormat soap_format("application/soap+xml");
  AFIAddress soap_addr(/* port = */ 9);
  soap.SetDataType("control");
  soap.SetTransportType("HTTP");
  soap.AddFormat(soap_format);
  soap.SetRxAddress(soap_addr);
  session->GetSessionList().push_back(soap);
#endif
  
  std::cout << "Session is " << *session << std::endl;

  mycall.SetSelfSessionDescription(session);
}

void MySIPCall::Publish(const std::string& event, const SIPMessageBody& body)
{
  SIPRequestMessage msg = BuildRequest("PUBLISH");

  msg.SetBody(body);

  SIPHeader event_header("Event", event);
  msg.SetHeader(event_header);
  SIPHeader expires_header("Expires", "3600");
  msg.SetHeader(expires_header);

  SendRequest(msg);  
}

/* print the information about the software */
static void print_software_info()
{
  cout << "SIP-Bluetooth Gateway";
#ifdef VERSION
  cout << " " << VERSION;
#endif
  cout << ", (c) 2005, Columbia University" << endl;
  cout << "Visit http://www.cs.columbia.edu/IRT/cinema for more info" << endl;
}

void usage(const std::string & command)
{
  const char *usage = "[options]\n"
    " -v          : Print version and exit\n"
    " -d category : verbose mode for the category. (see document)\n"
    "               all,sql,net,sdp,misc,...\n"
    " -o tracefile: put debug trace in this file.\n"
    " -p port     : listening SIP port, default is 5060.\n"
    " -i          : use dotted decimal for self address.\n"
    " -R uri      : SIP URI of the proxy/registrar, if any.\n"
    " -P uri      : SIP URI of the presence server, if any.\n"
    " -M uri      : SIP URI of the mapping server, if any.\n"
    " -m \"mac uri \" : MAC address to SIP URI mapping, if any.\n"
    "\n"
    "You can invoke the application with following options for example\n"
    " -d net -d app -p9000 -m \"00-0B-0D-06-9B-3B sip:user@cs.columbia.edu\"\n"
    ;

  cerr << "Usage: " << command << " "  << usage << endl;

  exit(1);
}


/* the server object */
MySIPUserAgent sip;

/* Main routine */
int main(int argc, char *argv[])
{
  

  /* assign the listening port */
  sip.SetListeningPort(9000);

  /* Set the server name */
  SIPLibrary::SetName("BluetoothGateway-Columbia-University");
  /* Initialize debug options */
  SIPLibrary::ErrorOn("stderr");
  /* Override SIPLib defaults here ... */
  SIPLibrary::SetThreadPoolSize(32); /* 32 is enough here */

  /* Print the software information */
  print_software_info();
        

  int c;
  
  while ((c = getopt(argc, argv, "hvd:o:p:iR:P:M:m:")) != EOF) {
    switch (c) {
      case 'v' :
#ifdef VERSION
        printf("sip-bluetooth version %s\n", VERSION);
#else
        printf("sip-bluetooth version unknown, not compiled with VERSION macro.\n");
#endif
        exit(0);
        break;
      case 'd':
        SIPLibrary::DebugOn();
        DebugAllowModule(optarg, 1);
        break;
      case 'o':
        SIPLibrary::DebugOn(optarg);
        break;
      case 'p':
        sip.SetListeningPort(atoi(optarg));
        break;
      case 'i' :
        SIPLibrary::UseNumericAddress(true);
        break;
      case 'R' :
	proxy = std::string(optarg);
	break;
      case 'P' : 
	presenceserver = std::string(optarg);
        break;
      case 'M' :
	mapping_server = std::string(optarg);
        break;
      case 'm' :
       {
	 list<string> nv = strtokenize(string(optarg), " ");
	 if (nv.size() < 2) {
	   cerr << "The -m option should have mac and user as \"mac user\"" << endl;
	   exit(1);
	 }
	 string mac = nv.front(); nv.pop_front();
	 string user = nv.front();
	 cout << "MAP " << mac << " => " << user << endl;
	 mac_users.insert(make_pair(mac, user));
       }
       break;
      case '?':
        /* no break */
      case 'h':
        usage(argv[0]);
        exit(1);
    } 
  }
  cout << "read params" << endl;

  SIPLibrary::Initialize();

  /* Start the listening server. */
  sip.StartServer();
  cout << "SIP " << sip << endl;

  handle_bluetooth();

  SIPLibrary::Reset();

  cout << "Exiting" << endl;
  return 0;
}

static void handle_bluetooth()
{
  BTManageAgent btm;
  list<string> added;
  list<string> removed;

  debug(DEBUG_MISC, "handle_bluetooth",
	"Activating bluetooth presence and registration\n");

  while(1){
    usleep(100000); /* wait for 100 ms */

    if (btm.Inquire(added, removed) == true) {
      /* something is added or removed */
      if (added.size() > 0) {
	/* something is added. */
	for (list<string>::iterator i = added.begin();
	     i != added.end(); ++i) {
	  handle_device_add((*i));
	}
      }
      if (removed.size() > 0) {
	/* something is removed. */
	for (list<string>::iterator i = removed.begin();
	     i != removed.end(); ++i) {
	  handle_device_remove((*i));
	}
      }
    }
  }
}

static void handle_device_add(const std::string& device)
{
  std::string server = "copenhagen.clic.cs.columbia.edu";

  cout <<" device detected is "<<device << endl;
  stringstream dev;
  dev << "sip:"<<device<<"@cs.columbia.edu";
  cout <<" device id created to be stored is "<<dev.str() << endl;
  string user;
  map<string,string>::iterator u_i = mac_users.find(dev.str());

  
    if (mapping_server.empty()) {
      error("handle_device_add", "No mapping server specified\n");
      return;
    }

  query_mapping_server(device);

  
}

static void handle_mapping_success(const std::string& mac_uri, 
				  const std::string& user)
{
  if (user != "") {
    mac_users.insert(make_pair(mac_uri, user));
    publish_user(user, true);
    register_user(user, user, true);
  }
}

static void handle_device_remove(const std::string& device)
{
	string user;
  cout <<" device to be removed is "<<device << endl;
  stringstream dev;
  dev << "sip:"<<device<<"@cs.columbia.edu";
  cout <<" device id  to be removed is "<<dev.str() << endl;
  map<string,string>::iterator u_i = mac_users.find(dev.str());


 user = (*u_i).second;

 publish_user(user, false);

 string from = string("sip:") + device + "@cs.columbia.edu";

 //register_user(user, user, false);

  
}

/* Format the publish message and send. */
static void query_mapping_server(const std::string& mac)
{

  /* The mapping server accepts address as sip:MAC@cs.columbia.edu */
  
  //IptelAddress to = mapping_server;
  string to = mapping_server;

  //to.SetUser(mac);
  
  MySIPCall *pub = new MySIPCall(sip);
  if (pub == NULL) {
    error("query_mapping_server", "new MySIPCall failed\n");
    return;
  }
  pub->SetProxyAddress(to);
  string remote = std::string("sip:")+std::string(mac)+std::string("@cs.columbia.edu");
  IptelAddress remoteurl = remote;
  remoteurl.SetScheme("sip");
  pub->SetRemoteAddress(remoteurl);
  string src = std::string("sip:")+std::string(mac)+std::string("@cs.columbia.edu");
  IptelAddress srcurl    = src;
  srcurl.SetScheme("sip");
  pub->SetSelfAddress(srcurl);

  pub->SetPreferredTransportType(SOCK_DGRAM);


  SIPMessageBody body;
  body.SetContentType("text/plain");
  body.SetContent("");

  pub->Publish("lookup", body);
  

}

static std::string format_publish_body(std::string user, std::string status) {
  return std::string(
      "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n")
    + "<impp:presence xmlns:impp=\"urn:ietf:params:xml:ns:pidf\"\n"
    + "      entity=\"pres:" + user + "\">\n"
    + "  <impp:tuple id=\"efe7284\">\n"
    + "    <impp:status>\n"
    + "      <impp:basic>" + status + "</impp:basic>\n"
    + "    </impp:status>\n"
    + "  </impp:tuple>\n"
    + "</impp:presence>\n";
}



/* Format and publish the user */
static void publish_user(const std::string& user, bool online)
{
  IptelAddress to = user;
  
  MySIPCall *pub = new MySIPCall(sip);
  if (pub == NULL) {
    error("query_mapping_server", "new MySIPCall failed\n");
    return;
  }

  pub->SetRemoteAddress(to);
  pub->SetSelfAddress(to);

  if (!presenceserver.empty()) {
    pub->SetProxyAddress(presenceserver);
  } 

  string status = online ? "pairing" : "non-pairing";

  SIPMessageBody body;
  body.SetContentType("application/pidf+xml");
  body.SetContent(format_publish_body(to.GetUserHost(), status));

  debug(DEBUG_MISC, "publish_user", "Body is %s\n",
	body.GetContent().c_str());

  pub->Publish("presence", body);
}



static std::string password;
static std::string login;

/* Format and register the user */
static void register_user(const std::string& user, 
			  const std::string& from,
			  bool online)
{
  IptelAddress to = user;
  IptelAddress contact;  

  /* Register with SIP server */
  SIPOutgoingRegister *reg = new SIPOutgoingRegister(sip);

  if (!proxy.empty()) {
    reg->SetProxyAddress(proxy);
  } 

  reg->SetSelfAddress(from);
  reg->SetRemoteAddress(to);

  contact = sip.BuildContactAddress(from);

  reg->SetContactAddress(contact);

  if (!login.empty()) {
    reg->SetAuthorization(login, password);
  }

  if (online == false) {
    //reg->Unregister();
  }
  else {
    reg->Register();
  }
}

