/* radio.cpp - Sipua tool handler.
 * 2000-2003 Columbia University. All rights reserved.
*/
#if defined(__sun__) && defined(__svr4__) /* Solaris */
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <sys/audioio.h>
#elif defined(__linux__)  /* Linux */
 #include <fcntl.h>
 #include <sys/soundcard.h>
 #include <sys/stat.h>
 #include <sys/ioctl.h>
#else
#if __GNUC__
#warning "audio device interface will not work on your platform"
#else
#pragma message("warning: audio device interface will not work on your platform")
#endif
#endif

#include <string>         /* std::string */
#include "rtpsession.h"   /* RTP::Session */
#include "afiaddress.h"   /* AFIAddress */
#include "radio.h"        /* class definition AudioTool */
#include "siplibrary.h"
#include "string-util.h"

extern "C" {
#include <sys/types.h>
#include <signal.h>

#include <pthread.h>
#include "pthread-util.h"
#include "error.h"
#include "timespec.h"
#include "sysdep.h"
}

/* Random32 */
extern "C" unsigned int random32(int type);

using namespace std;
using namespace RTP;

/////////////////////////// Macros //////////////////////////////////

#define DEFAULT_THRESHOLD 0 // Change this to 500000


////////////////////////// External functions ///////////////////////
extern std::list<std::string> strtokenize(const std::string & str,
                                          const std::string & seperator);
////////////////////////// Method definitions ///////////////////////

/* Construct */
AudioTool::AudioTool(const std::string path)
: RTP::RTPSession()
{
  active = false;
  tool = path;
}

///destroy the audio tool associated with this call.
/// does RTP library send BYE automatically? XXX
AudioTool::~AudioTool()
{
  if (active)
    stop();
  active = false;
}

/* set parameters and start RTP session */
bool AudioTool::start_rtp()
{
  /* Internal tool */
  AFIAddress addr = get_destination();
  if (addr.get_port() == 0) {
    addr.set_any_host();
    addr.set_port(0);
    set_destination(addr);
  }
  
  /* Set default payload type as 0 */
  std::pair<payload_type_t,double> payloadtype
    = get_payload_type();
  if (payloadtype.first > 127) { /* Since this is unsigned int8 */
    set_payload_type(0, 8000.);
  }
  
  /* Set the cname in RTP session */
  if (cname.empty()) {
    cname = getenv("LOGNAME") ? getenv("LOGNAME") :
      (getenv("USER") ? getenv("USER") : "sipua");
    cname += "@";
    cname.append(AFIAddress::get_local_hostname());
  }
  set_local_member_cname(cname);
  
  /* open the RTP session */
  try {
    open();
  } 
  catch (exception &e) {
    NOEXCEPT_DUMMYDECL(exception e);
    error("start", 
          "Could not start the audio tool, RTPSession::open failed: %s\n",
          e.what() ? e.what() : "NULL");
    return false;
  }

  return true;
}

/** Do not kill the device but ignore any input from the device. */
bool AudioTool::set_microphone_volume(int volume)
{
  mic_volume = volume;
  return true;
}

/** Ignore any output to the device */
bool AudioTool::set_speaker_volume(int volume)
{
  speaker_volume = volume;
  return true;
}

bool AudioTool::mute_input(bool mute)
{
  mic_mute = mute;
  return true;
}

bool AudioTool::mute_output(bool mute)
{
  speaker_mute = mute;
  return true;
}

bool AudioTool::is_silence(RTP::payload_type_t pt,
                           std::string & payload) const
{
#if 1
  return false;
#else

#define SQR(x) (x)*(x)
  int i, n;
  double sum_x, avg_x;

  n = payload.length();

  /* Following portion of the code is copied from HW2 */
  for (sum_x=0, i=0; i < n; i++) { 
    sum_x += pcmu_l16(payload[i]); 
  }

  avg_x = sum_x/n;
  for (sum_x=0, i=0; i < n; i++) { 
    sum_x += 
      SQR(pcmu_l16(payload[i]) - avg_x); 
  }
  avg_x = sum_x/n;

  return (avg_x < silence_threshold);
#endif
}

////////////////////////// file variables ///////////////////////////

/* static variables */
static std::list<DeviceAudioTool *> listener;
static in_port_t port_base = 10000;
static pthread_t playout_threadid = 0;
static pthread_t record_threadid = 0;
static bool device_started = false;
static pthread_mutex_t device_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t  device_cond  = PTHREAD_COND_INITIALIZER;
static int audio_fp = -1;
static int device_encoding = 1;
static int device_sample_rate = 8000;
static int device_sample_size = 1;
static int record_interval    = 20;
static int playout_interval   = 20;

///class method to open the audio device.
static bool open_device();

/// class method to close an audio device
static bool close_device();

///class method to start the thread for device.
static bool start_device();

///class method to stop the thread for the device.
static bool stop_device();

///check if the device is open.
static bool is_device_open() {return audio_fp != -1; }

///check if device is started.
static bool is_device_started() {return device_started; }

///set the device started boolean.
static void set_device_started(bool s) {device_started = s; }

///internal thread for playout and mixing.
static void *playout_thread(void *); 

///internal thread for recording and distribution to
/// all the listener AudioTool objects.
static void *record_thread(void *);

///Install a new listener.
static bool install_listener(DeviceAudioTool *l);

///Un-Install a listener.
static bool uninstall_listener(DeviceAudioTool *l);

///Has a particular listener.
static bool has_listener(DeviceAudioTool *l);

///Has any listener.
static bool has_listeners();

///get a packet for playout.
static int get_packet_playout(unsigned char *buffer, int size);


/**Start the audio tool. This may open the 
   audio device or invoke the external tool. */
bool DeviceAudioTool::start()
{
  if (active)
    return true;  /* already started */

  if (!start_rtp()) {
    return false;
  }

  /* set the tool as being active now */
  active = true;
  
  install_listener(this);
  if (!is_device_open()) {
    open_device();
    start_device();
  }

  return active;
}

/**Stop the audio tool. This may close the audio device
   or kill the external tool. */
bool DeviceAudioTool::stop()
{
  /* Kill the threads */
  if (active) {
    /* Close RTP::Session */
    close();
    
    /* detach from device */
    uninstall_listener(this);
    
    /* If this is the last listener then close the device */
    if (is_device_open() && 
        !has_listeners()) {
      stop_device();
      close_device();
    }
    active = false;
  }

  return true;
}

/* Incoming audio packet from remote using RTP */
void DeviceAudioTool::receive(const RTP::RTPMemberPtr& sender,
                     RTP::rtp_timestamp_t ts, 
                     RTP::payload_type_t pt,
                     bool marker, 
                     const std::string& payload)
{
  /* Just play the audio to device */
  play_audio(ts, pt, marker, payload);
}

/** Call back on incoming audio input. */
bool DeviceAudioTool::on_audio_recorded(RTP::rtp_timestamp_t ts,
                                  RTP::payload_type_t pt,
                                  bool marker,
                                  const std::string& payload)
{
  /* send the audio to remote using RTP */
  if (is_open()) {
    try {
      send(ts, marker, payload);
      return true;
    }
    catch (exception&) {
      return false;
    }
  }
  else {
    /* not open */
    return false;
  }
}

/** Send audio for playback */
bool DeviceAudioTool::play_audio(RTP::rtp_timestamp_t ts,
                           RTP::payload_type_t pt,
                           bool marker,
                           const std::string& payload)
{
  /* Apply the delay adjustment algorithm */

  return true;
}

bool DeviceAudioTool::rfc2833_send(const std::string& digits)
{
  for (unsigned i=0; i<digits.size(); i++) { 
    char digit = digits[i];
    std::string buffer;

    if (digit >= '0' && digit <= '9') {
      buffer[0] = digit - '0';
    }
    else if (digit == '*') {
      buffer[0] = 10;
    }
    else if (digit == '#') {
      buffer[0] = 11;
    }
    else {
      continue;
    }

    buffer[1] = (unsigned char) 0x94;
    buffer[2] = (unsigned char) 0x07;
    buffer[3] = (unsigned char) 0xd0;

    send(time_stamp /* timestamp = 0 ?? XXX */,
         false /* marker is false */, 
         buffer);
  }

  return true;
}

/* static functions to handle the local device */
static bool open_device()
{
  bool ret = false;

  if (is_device_open()) {
    return true; /* already open, close first. */
  }

#if defined(__sun__) && defined(__svr4__)
  /* Solaris audio */
  debug(DEBUG_MISC, "open_device", 
                    "opening /dev/audio\n");

  mode_t mode = O_RDWR;
  if ((audio_fp = open("/dev/audio", mode | O_NDELAY | O_NONBLOCK)) < 0) {
    error("open_device", "Error opening /dev/audio\n");
    return false;
  }

  /* Set the speaker and microphone configuration */
  int audioctl_fp = -1;
  if ((audioctl_fp = open("/dev/audioctl", O_RDWR | O_NDELAY)) < 0) {
    error("open_device", "Error opening /dev/audioctl\n");
  }
  else {
    audio_info_t ai;
    
    /* Set speaker and microphone ports */
    AUDIO_INITINFO(&ai);
    ai.record.port        = AUDIO_MICROPHONE;
    ai.play.port          = AUDIO_SPEAKER;
    ai.record.sample_rate = device_sample_rate;
    ai.record.channels    = 1;
    ai.record.encoding    = device_encoding; /* Mu law */
    ai.play.sample_rate   = device_sample_rate;
    ai.play.channels      = 1;
    ai.play.encoding      = device_encoding;

    /* 
     * somehow the audio info does not seem to be set,
     * for encoding. Use "cp bg.au /dev/audio" before
     * running this program to set the encoding.
     */ 
    ioctl(audioctl_fp, AUDIO_SETINFO, &ai);
    close(audioctl_fp);
  }
  ret = true;

#elif defined(__linux__)
  /* Linux audio */
  debug(DEBUG_MISC, "open_device", 
                    "opening /dev/audio\n");

  mode_t mode = O_RDWR;
  if ((audio_fp = open("/dev/audio", mode | O_NDELAY | O_NONBLOCK, NULL)) < 0) {
    error("open_device", "Error opening /dev/audio\n");
    return false;
  }

  int format = AFMT_MU_LAW, stereo=0, speed=device_sample_rate, duplex=1;

  if(ioctl(audio_fp, SNDCTL_DSP_SETFMT, &format) == -1){
    error("open_device", "Can not set format to AFMT_MU_LAW\n");
    close(audio_fp);
    return false;
  }
  if(ioctl(audio_fp, SNDCTL_DSP_STEREO, &stereo) == -1){
    error("open_device", "Can not set mode to not-stereo\n");
    close(audio_fp);
    return false;
  }
  if(ioctl(audio_fp, SNDCTL_DSP_SPEED, &speed) == -1){
    error("open_device", "Can not set speed to 8000\n");
    close(audio_fp);
    return false;
  }
  if(ioctl(audio_fp, SNDCTL_DSP_SETDUPLEX, &duplex) == -1){
    error("open_device", "Can not set duplex=true\n");
    close(audio_fp);
    return false;
  }

  ret = true;

#else
  error("open_device", 
        "The audio device interface on this platform is not implemented\n");
#endif

  return ret;
}

static bool close_device()
{
  if (!is_device_open()) {
    /* Already closed */
    return true;
  }

#if defined(__sun__) && defined(__svr4__) || defined(__linux__)
  close(audio_fp);
#else
  error("close_device", 
        "The audio interface on this platform is not implemented\n");
#endif

  audio_fp = -1; /* reset audio file pointer */
  return true;
}

static bool start_device()
{
  if (is_device_started()) {
    return true; /* already started */
  }

  pthread_mutex_lock(&device_mutex);
  set_device_started(true);

  /* 
   * Setup threads to listen to incoming packets and send outgoing
   * packets 
   */
  int  e;

  /* Recv thread */
  if ((e = pthread_create(&playout_threadid, NULL, 
    playout_thread, (void *)NULL)) != 0) {
    error("start_device", "pthread_create_detached failed");
    pthread_mutex_unlock(&device_mutex);
    return false;
  }
  
  if ((e = pthread_create(&record_threadid, NULL, 
    record_thread, (void *)NULL)) != 0) {
    error("start_device", "pthread_create_detached failed");
    pthread_mutex_unlock(&device_mutex);
    stop_device();
    return false;
  }

  pthread_mutex_unlock(&device_mutex);
  return true;
}

static bool stop_device()
{
  if (!is_device_started()) {
    /* Not started. no need to close */
    return false;
  }

  pthread_mutex_lock(&device_mutex);
  set_device_started(false); /* signal to threads to terminate. */
  pthread_cond_broadcast(&device_cond);

  void *ret = NULL;
  pthread_join(playout_threadid, &ret);
  ret = NULL;
  pthread_join(record_threadid, &ret);

  pthread_mutex_unlock(&device_mutex);
  return true;
}

static bool has_listener(DeviceAudioTool * tool)
{
  bool ret = false;
  pthread_mutex_lock(&device_mutex);
  std::list<DeviceAudioTool *>::const_iterator iter;
  for (iter = listener.begin(); iter != listener.end(); ++iter) {
    if ((*iter) == tool) {
      ret = true;
      break;
    }
  }
  pthread_mutex_unlock(&device_mutex);
  return ret;
}

static bool has_listeners()
{
  return !listener.empty();
}

static bool install_listener(DeviceAudioTool *tool)
{
  pthread_mutex_lock(&device_mutex);
  listener.push_back(tool);
  pthread_mutex_unlock(&device_mutex);
  return true;
}

static bool uninstall_listener(DeviceAudioTool *tool)
{
  pthread_mutex_lock(&device_mutex);
  std::list<DeviceAudioTool *>::iterator iter;
  for (iter = listener.begin(); iter != listener.end(); ) {
    if ((*iter) == tool) {
      iter = listener.erase(iter);
    }
    else {
      ++iter;
    }
  }
  pthread_mutex_unlock(&device_mutex);
  return true;
}

static void *record_thread(void *)
{
  unsigned char buffer[8192];
  int size, len;
  std::list<DeviceAudioTool *>::iterator iter;

  while (is_device_started()) {

    /* packet size is samplesize * sampling rate * interval in seconds */
    size = (device_sample_size * device_sample_rate * record_interval / 1000);

    if (size > (int) sizeof(buffer))   
      size = sizeof(buffer);

    len = read(audio_fp, buffer, size);
    if (len > 0 && len < size) {
      memset(&buffer[len], 0x7f, size - len);
      
      SIPLibrary::SleepInMillisecond((size - len) / 8);
    }
    else {
      SIPLibrary::SleepInMillisecond(size / 8);
    }

    pthread_mutex_lock(&device_mutex);

    for (iter = listener.begin(); iter != listener.end(); ++iter) {
      if (*iter != NULL) {
        (*iter)->increment_timestamp(device_sample_rate * record_interval / 1000);

        std::string payload((const char *)buffer, size);
        
        /* transcode now */
        RTP::payload_type_t pt = device_encoding == 1? 0 : -1;

        bool silence = (*iter)->is_silence(pt, payload);

        if ((*iter)->is_silence() && !silence) {
          (*iter)->set_silence(false);
        }
        if (!silence) {

          /* call the callback */
          (*iter)->on_audio_recorded(pt, 
                                    (*iter)->get_timestamp(),
                                    (*iter)->is_silence(),
                                    payload);
        }
        else {
          (*iter)->set_silence(true);
        }
      }
    }

    pthread_mutex_unlock(&device_mutex);
  }

  return NULL;
}

static void *playout_thread(void *)
{
  unsigned char buffer[8192];

  pthread_mutex_lock(&device_mutex);
  
  struct timeval initialtime;
  struct timespec next;

  gettimeofday(&initialtime, NULL);
  double total_duration = 0.0;
  int size;
  
  /* each loop should ideally iterate for playout_interval.
  * otherwise its a overload condition */

  while (is_device_started()) {
    next = timespec_delay(initialtime, total_duration);
    total_duration += playout_interval / 1000.;
    size = playout_interval * device_sample_rate * device_sample_size / 1000;
    pthread_cond_timedwait(&device_cond, &device_mutex, &next);
    if (!is_device_started()) {
      break; /* stopped */
    }

    if (get_packet_playout(buffer, size) >= 0) {
      write(audio_fp, buffer, size);
    }
  }

  pthread_mutex_unlock(&device_mutex);

  return NULL;
}

/** get data for playout */
static int get_packet_playout(unsigned char *buffer, int size)
{
  return 0;
}

/***************** External tool specific implementation ************/

bool ExternalAudioTool::start()
{
  if (active)
    return true;  /* already started */

  std::string args;
  
  AFIAddress remoteaddress = get_destination();
  AFIAddress selfaddress   = get_receive_address();
  
  args += remoteaddress.get_numeric_address();
  args += "/";
  if (selfaddress.get_port() != 0 &&
      selfaddress.get_port() && remoteaddress.get_port()) { 
    args += selfaddress.get_port();
    args += "/";
  }
  args += remoteaddress.get_port();
  
  active = start_external_tool(args);
  /* Do not install as a listener */

  return active;
}

bool ExternalAudioTool::stop()
{
  /* Kill the threads */
  if (active) {
    stop_external_tool();
    active = false;
  }

  return true;
}

/* Internal protected methods */
bool ExternalAudioTool::start_external_tool(std::string args)
{
  std::string command = tool;
#ifndef WIN32
  /* for unix systems use execv */
  command += " ";
  command += args;

  std::list<std::string> p = strtokenize(command, " \t");

  if (p.size() < 1) {
    return false;
  }

  std::list<std::string>::iterator i;
  i = p.begin();

  const char *filename = (*i).c_str();

  char *argv[10];
  unsigned j;
  for (j=0; i != p.end() && j<10; ++i, j++) {
    argv[j] = (char *) (*i).c_str();
  }
  argv[j] = 0;

  int retval;
  switch ((retval = fork())) {
  case 0:
    execv(filename, argv);
  case -1:
    return false; /* some fork error */
  default:
    pid = retval;  /* store the process id */
    return true;
  }

#else
  /* for windows system use spawnlp */
  pid = _spawnlp(_P_NOWAIT, 
                 command.c_str(), 
                 command.c_str(), 
                 args.c_str(), NULL);
  return true;
#endif
}

bool ExternalAudioTool::stop_external_tool()
{
#ifndef WIN32
  /* Unix systems use Kill */
  if (pid > 0) {
    if (kill(pid, SIGTERM) < 0) {
      error("StopExternalTool", "can not kill process id %d\n", pid);
    }
  }
#endif
  /* 
   * We assume there is a GUI that allows terminating for
   * a windows system 
   */
  return true;
}

/********************************* FILE Audio Tool ******************/

/** Construct a file audio interface */
FileAudioTool::FileAudioTool(const std::string read_from, 
                             const std::string write_to,
                             bool r)
  : AudioTool(), read_file(read_from), write_file(write_to)
{

  pthread_mutex_init(&read_mutex, NULL);
  pthread_mutex_init(&write_mutex, NULL);

  read_fp = NULL;
  write_fp = NULL;
  repeat = r;
}

/** Start the tool */
bool FileAudioTool::start()
{
  bool ret = false;

  if (active)
    return true; /* already started */

  if (!start_rtp()) {
    return false;
  }

  if (start_read()) {
    ret = true;
  }

  if (start_write()) {
    ret = true;
  }

  if (!ret) {    
    close();
    stop_read();
    stop_write();
  }

  return ret;
}

/** Stop the tool */
bool FileAudioTool::stop()
{
  if (active) {
    /* Close RTP::Session */
    close();
    
    active = false;
  }

  stop_read();
  stop_write();

  return true;
}

/** The read thread to read and send using RTP */
void *file_read_thread(void *p)
{
  FileAudioTool *tool = (FileAudioTool *)p;
  if (!tool) {
    error("file_read_thread", "NULL parameter\n");
    return NULL;
  }

  pthread_mutex_lock(&tool->read_mutex);

  unsigned char buffer[160];
  int len, size = sizeof(buffer);
  bool first_time = true;
  RTP::rtp_timestamp_t ts = random32(time(0));

  do {
    len = fread(buffer, 1, size, tool->read_fp);

    if (len <= 0 || len > size) {
      if (tool->repeat && feof(tool->read_fp)) {
        fseek(tool->read_fp, 0, SEEK_SET);
        continue;
      }
      break;
    }

    memset(&buffer[len], 0x7f, size - len); /* pad with silence */

    try {
      tool->send(ts, first_time, std::string((const char *) buffer, size));      
      first_time = false;
    }
    catch (std::exception& e) {
      NOEXCEPT_DUMMYDECL(std::exception e);
      error("file_read_thread", "Can not send RTP: %s\n", e.what());
    }

    ts += size; /* increment timestamp */

    pthread_mutex_unlock(&tool->read_mutex);
    SIPLibrary::SleepInMillisecond(size / 8); /* 160/8 = 20 ms packets */
    pthread_mutex_lock(&tool->read_mutex);

  } while (tool->read_fp);


  /* Do not close file here, its closed in stop_read() */    
  pthread_mutex_unlock(&tool->read_mutex);

  return NULL;
}


/** Start the read module */
bool FileAudioTool::start_read()
{
  bool ret = true;
  int  e;

  pthread_mutex_lock(&read_mutex);

  debug(DEBUG_MISC, "start_read", "opening %s\n", read_file.c_str());

  if (read_file.empty()) {
    error("start_read", "Read file is empty\n");
    ret = false;
  }
#ifdef WIN32
  else if ((read_fp = fopen(read_file.c_str(), "rb")) < 0) 
#else
  else if ((read_fp = fopen(read_file.c_str(), "r")) < 0) 
#endif
  {
    error("start_read", "Error opening %s\n", read_file.c_str());
    ret = false;
  }
  else  if ((e = pthread_create(&read_thread, NULL, 
                                file_read_thread, (void *)this)) != 0) {
    error("start_read", "pthread_create failed\n");
    fclose(read_fp);
    ret = false;
  }

  if (!ret) {
    read_fp = NULL;
  }

  pthread_mutex_unlock(&read_mutex);
    
  return ret;
}

/** Stop the read thread */
bool FileAudioTool::stop_read()
{
  pthread_mutex_lock(&read_mutex);

  if (read_fp) {
    void *val = NULL;

    fclose(read_fp);
    read_fp = NULL;

    pthread_mutex_unlock(&read_mutex);
    pthread_join(read_thread, &val);
    pthread_mutex_lock(&read_mutex);
  }

  pthread_mutex_unlock(&read_mutex);

  return true;
}

/** Start the write file */
bool FileAudioTool::start_write()
{
  bool ret = true;

  pthread_mutex_lock(&write_mutex);

  debug(DEBUG_MISC, "start_write", "opening %s\n", write_file.c_str());

  if (write_file.empty()) {
    error("start_write", "Write file is empty\n");
    ret = false;
  }
#ifdef WIN32
  else if ((write_fp = fopen(read_file.c_str(), "wb")) < 0) 
#else
  else if ((write_fp = fopen(read_file.c_str(), "w")) < 0) 
#endif
  {
    error("start_write", "Error opening %s\n", write_file.c_str());
    ret = false;
  }

  pthread_mutex_unlock(&write_mutex);

  if (!ret) {
    write_fp = NULL;
  }

  return ret;  
}

/** Stop write */
bool FileAudioTool::stop_write()
{
  pthread_mutex_lock(&write_mutex);
  if (write_fp) {
    fclose(write_fp);
    write_fp = NULL;
  }
  pthread_mutex_unlock(&write_mutex);

  return true;
}

/* Incoming audio packet from remote using RTP */
void FileAudioTool::receive(const RTP::RTPMemberPtr& sender,
                     RTP::rtp_timestamp_t ts, 
                     RTP::payload_type_t pt,
                     bool marker, 
                     const std::string& payload)
{
  /* Just write to the file. XXX we can not lock write_mutex
  * since that may cause deadlock due to order of locking with
  * session_lock */
  if (write_fp) {
    unsigned int len;
    len = fwrite(payload.c_str(), 1, payload.length(), write_fp);
    if (len != payload.length()) {
      error("receive", "fwrite failed (%d != %d)\n", len, payload.length());
    }
  }
}

/*********************** Loopback audio interface *********************/

LoopbackTool::LoopbackTool() 
  : AudioTool()
{
  pthread_mutex_init(&mutex, NULL);
  pthread_cond_init(&cond, NULL);
}

LoopbackTool::~LoopbackTool()
{
  if (!active) {
    stop();
  }

  pthread_mutex_destroy(&mutex);
  pthread_cond_destroy(&cond);

  while(!queue.empty()) {
    LoopbackTool::mypacket_t *packet = queue.front();
    delete packet;
    queue.pop_front();
  }
}

/* Incoming audio packet from remote using RTP */
void LoopbackTool::receive(const RTP::RTPMemberPtr& sender,
                           RTP::rtp_timestamp_t ts, 
                           RTP::payload_type_t pt,
                           bool marker, 
                           const std::string& payload)
{
  if (!terminate) {
    pthread_mutex_lock(&mutex);
    queue.push_back(new LoopbackTool::mypacket_t(payload, marker, ts));
    pthread_cond_broadcast(&cond);
    pthread_mutex_unlock(&mutex);
  }  
}

void *loopback_thread(void *p)
{
  LoopbackTool *tool = (LoopbackTool *)p;
  LoopbackTool::mypacket_t *packet;
  
  if (tool == NULL) {
    return NULL;
  }

  pthread_mutex_lock(&tool->mutex);

  while (true) {
    pthread_cond_wait(&tool->cond, &tool->mutex);
    if (tool->terminate)
      break;

    if (!tool->queue.empty()) {
      packet = tool->queue.front();
      tool->queue.pop_front();

      if (packet != NULL) {
        pthread_mutex_unlock(&tool->mutex);

        try {          
          tool->send(packet->ts, packet->marker, packet->payload);
        }
        catch (std::exception& e) {
          NOEXCEPT_DUMMYDECL(std::exception e);
          error("loopback_thread", "Can not send RTP: %s\n", e.what());
        }  
        delete packet;

        pthread_mutex_lock(&tool->mutex);
      }
    }
  }

  pthread_mutex_unlock(&tool->mutex);

  return NULL;
}

/** Start the tool */
bool LoopbackTool::start()
{
  int e;

  if (active)
    return true; /* already started */

  if (!start_rtp()) {
    return false;
  }

  terminate = false;
  if ((e = pthread_create(&loopback_threadid, NULL, 
                          loopback_thread, (void *)this)) != 0) {
    error("start", "pthread_create failed\n");
    close();
    terminate = true;
    return false;
  }

  return true;
}

/** Stop the tool */
bool LoopbackTool::stop()
{
  void *val = NULL;

  if (active) {
    /* Close RTP::Session */
    close();
    
    active = false;
  }

  pthread_mutex_lock(&mutex);
  if (!terminate) {
    terminate = true;
    pthread_cond_broadcast(&cond);
    pthread_mutex_unlock(&mutex);
    pthread_join(loopback_threadid, &val);
  }
  else {
    pthread_mutex_unlock(&mutex);
  }

  return true;
}

LoopbackTool::mypacket_t::mypacket_t(const std::string& p, 
                                     bool m, RTP::rtp_timestamp_t t)
{
  payload = p;
  marker = m;
  ts = t;
}
