#ifndef RADIO_H_
#define RADIO_H_

#include <string>
#include <list>
#include <exception>

#include <pthread.h>
#include <stdio.h>

#include "rtpsession.h"  /* RTPSession */
#include "afiaddress.h"  /* AFIAddress */

/**
 * AudioTool class is a RTP interface to audio device.  It represents
 * an RTP session for audio and also interfaces with the physical
 * audio device on Sun and Linux. This AudioTool class automatically
 * connects the RTP library with the audio device without the
 * application intervention.  Multiple instances of the audio tool
 * call be created, however will refer to the same underlying device.
 * The playback from all audio tools to the same device are mixed
 * before playing. The input from microphone is given to all the audio
 * tool interfaces associated with a device. This is an handy audio
 * device multiplexer, mixer.  If an external audio tool like vat or
 * rat is specified then each instance of audiotool will invoke one
 * such external tool. If the developer wants to get the raw audio
 * from and to the RTP or device then he needs to create a derived
 * class of Audio tool and overwrite the call back methods.
 */
class AudioTool : public RTP::RTPSession
{
public :
  /** Create a new audio tool connection for RTP session.  Use
   * internal PCMU based audio tool. Use the specified audio tool
   * (e.g., vat, rat) if path is present. */
  AudioTool(const std::string path = "");

  ///destroy the audio tool associated and RTP session.
  virtual ~AudioTool();

  /**Start the audio tool. This may open the 
     audio device or invoke the external tool. */
  virtual bool start() {return false;}

  /**Stop the audio tool. This may close the audio device
     or kill the external tool. */
  virtual bool stop() {return false;}

  /** Do not kill the device but ignore any input from the device. */
  virtual bool mute_input(bool mute = true);

  /** Ignore any output to the device */
  virtual bool mute_output(bool mute = true);

  /** set the microphone volume */
  virtual bool set_microphone_volume(int v);

  /** Set the speaker volume */
  virtual bool set_speaker_volume(int v);

  /* 
   * application can use RTP::Session methods like get_destination,
   * set_destination, set_receive_address, get_receive_address, send,
   * etc.  
   */

  ///send DTMF using RFC2833
  virtual bool rfc2833_send(const std::string & digits) {return false; }

  /** silence detection */
  virtual bool is_silence(RTP::payload_type_t pt,
                          std::string & payload) const;

  /** Is this a silence period */
  bool is_silence() const { return silence; }

  /** Set the silence period */
  void set_silence(bool s = true) {silence = s; }

  /** Get the current timestamp */
  RTP::rtp_timestamp_t get_timestamp() const { return time_stamp; }

  /** Increment the timestamp based on the clock ticks */
  void increment_timestamp(unsigned long delta) {time_stamp += delta; }

protected:
  
  /** Start the RTP session after setting the parameters */
  virtual bool start_rtp();

  /* member variables */
  bool active;

  std::string tool;

  /* other device parameters */
  int speaker_volume;  /* 0 to 255 */
  int mic_volume;      /* 0 to 255 */
  bool speaker_mute;
  bool mic_mute;

  /* Canonical name for this session */
  std::string cname;

  /* is incremented by the record thread */  
  RTP::rtp_timestamp_t time_stamp; 
  /* silence and marker bits */
  bool silence;
};

/**
 * This subclass implements the device specific methods for actual
 * audio device 
 */
class DeviceAudioTool : public AudioTool
{
 public:
  /** Construct a new device object */
  DeviceAudioTool(const std::string path="") : AudioTool(path) { };

  /** Destroy the device object */
  virtual ~DeviceAudioTool() { };

  /**Start the audio tool. This may open the audio device */
  virtual bool start();

  /**Stop the audio tool. This may close the audio device */
  virtual bool stop();

  /* Actually send RFC2833 style DTMF packet */
  virtual bool rfc2833_send(const std::string & digits);

  /* Incoming audio packet from remote using RTP */
  virtual void receive(const RTP::RTPMemberPtr& sender,
                       RTP::rtp_timestamp_t ts, 
                       RTP::payload_type_t pt,
                       bool marker, 
                       const std::string& payload);

  /** Call back on incoming audio input. */
  virtual bool on_audio_recorded(RTP::rtp_timestamp_t ts,
                                 RTP::payload_type_t pt,
                                 bool marker, const std::string & payload);

  /** Send audio for playback */
  virtual bool play_audio(RTP::rtp_timestamp_t ts,
                          RTP::payload_type_t pt,
                          bool marker, const std::string & payload);
};

/**
 * This sub-class implements the device specific methods for
 * external audio tool.
 */
class ExternalAudioTool : public AudioTool
{
 public:
  /** Construct a new external audio tool with specified path. */
  ExternalAudioTool(const std::string path) : AudioTool(path) { pid = 0; };

  /** Destroy the tool */
  virtual ~ExternalAudioTool() { };

  /**Start the audio tool. This may start the external application */
  virtual bool start();

  /**Stop the audio tool. This may kill the external application */
  virtual bool stop();

 protected:
  /** start an external tool */
  bool start_external_tool(std::string args);

  /** start an external tool */
  bool stop_external_tool();

  int pid;  /* process id of external tool if tool != "" */
};

/**
 * A file interface that reads from a file and writes into a file.
 */
class FileAudioTool : public AudioTool
{
 public:
  /** Construct a file audio interface */
  FileAudioTool(const std::string read_from, const std::string write_to,
                bool repeat = true);
  
  /** destroy the file audio tool */
  ~FileAudioTool() { stop(); }

  /** Start the tool */
  virtual bool start();

  /** Stop the tool */
  virtual bool stop();

  /* Incoming audio packet from remote using RTP */
  virtual void receive(const RTP::RTPMemberPtr& sender,
                       RTP::rtp_timestamp_t ts, 
                       RTP::payload_type_t pt,
                       bool marker, 
                       const std::string& payload);

 protected:
  /** Start the read thread */
  bool start_read();
  /** Stop the read thread */
  bool stop_read();

  /** Open the write file. */
  bool start_write();
  /** Close the write file */
  bool stop_write();

  /** The read file name for input (microphone substitute) */
  std::string read_file;

  /** The write file name for output (speaker substitute) */
  std::string write_file;

  /** read thread */
  pthread_t read_thread;
  pthread_mutex_t read_mutex;
  FILE *read_fp;
  bool repeat;

  /** write thread */
  pthread_mutex_t write_mutex;
  FILE *write_fp;

  friend void *file_read_thread(void *p);

};

/**
 * The loopback audio tool just loops back the received audio to send.
 */
class LoopbackTool : public AudioTool
{
 public:
  /** Construct an interface */
  LoopbackTool();
  
  /** destroy the file audio tool */
  ~LoopbackTool();

  /** Start the tool */
  virtual bool start();

  /** Stop the tool */
  virtual bool stop();

  /* Incoming audio packet from remote using RTP */
  virtual void receive(const RTP::RTPMemberPtr& sender,
                       RTP::rtp_timestamp_t ts, 
                       RTP::payload_type_t pt,
                       bool marker, 
                       const std::string& payload);

 protected:
  pthread_t loopback_threadid;
  pthread_mutex_t mutex;
  pthread_cond_t cond;
  bool terminate;

  class mypacket_t {
  public:
    mypacket_t(const std::string& p, bool m, RTP::rtp_timestamp_t ts);

    std::string payload;
    bool marker;
    RTP::rtp_timestamp_t ts;
  };

  std::list<mypacket_t *> queue;

  friend void *loopback_thread(void *);
};

#endif /* RADIO_H_ */

